from django.test import TestCase, override_settings
from django.urls import reverse

from django_hawk.tests.test_views import DjangoHawkViewTests, hawk_auth_sender


class DjangoViewTests(DjangoHawkViewTests, TestCase):
    view_name = "test_simple_view"


@override_settings(
    DJANGO_HAWK={
        "HAWK_INCOMING_ACCESS_KEY": "some-id",
        "HAWK_INCOMING_SECRET_KEY": "some-secret",
    }
)
class CustomHeaderViewTests(TestCase):
    def _get(self):
        url = "http://testserver" + reverse("custom_header_view")
        sender = hawk_auth_sender(url=url)

        return self.client.get(
            url,
            content_type="",
            HTTP_AUTHORIZATION=sender.request_header,
            HTTP_X_FORWARDED_FOR="1.2.3.4, 123.123.123.123",
        )

    def test_custom_header_is_in_response(self):
        response = self._get()

        self.assertEqual(response.status_code, 200)
        self.assertTrue("Content-Type" in response)
        self.assertTrue("Server-Authorization" in response)
        self.assertEqual(response["Server-Authorization"], "custom-header-value")



import os

import django.utils.crypto

TESTS_PATH = os.path.dirname(os.path.abspath(__file__))

SECRET_KEY = django.utils.crypto.get_random_string(50)

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
            ],
        },
    },
]

DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": "db.sqlite3"}}

MEDIA_ROOT = os.path.join(TESTS_PATH, "media")

STATIC_ROOT = os.path.join(TESTS_PATH, "static")

ROOT_URLCONF = "django_hawk.tests.urls"

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django_hawk.middleware.HawkResponseMiddleware",
]



import datetime

import mohawk
from django.test import override_settings
from django.urls import reverse
from freezegun import freeze_time


def hawk_auth_sender(
    url: str,
    key_id: str = "some-id",
    secret_key: str = "some-secret",
    method: str = "GET",
    content: str = "",
    content_type: str = "",
):
    credentials = {
        "id": key_id,
        "key": secret_key,
        "algorithm": "sha256",
    }
    return mohawk.Sender(
        credentials,
        url,
        method,
        content=content,
        content_type=content_type,
    )


class DjangoHawkViewTests:
    view_name: str = ""

    def get_path(self) -> str:
        return reverse(self.view_name)

    def get_url(self) -> str:
        return "http://testserver" + self.get_path()

    @override_settings(
        DJANGO_HAWK={
            "HAWK_INCOMING_ACCESS_KEY": "some-id",
            "HAWK_INCOMING_SECRET_KEY": "some-secret",
        }
    )
    def test_empty_object_returned_with_authentication(self):
        """
        If the Authorization and X-Forwarded-For headers are correct, then
        the correct, and authentic, data is returned
        """

        url = self.get_url()
        sender = hawk_auth_sender(url=url)
        response = self.client.get(
            url,
            content_type="",
            HTTP_AUTHORIZATION=sender.request_header,
            HTTP_X_FORWARDED_FOR="1.2.3.4, 123.123.123.123",
        )

        self.assertEqual(response.status_code, 200)
        self.assertTrue("Content-Type" in response)
        self.assertTrue("Server-Authorization" in response)

    @override_settings(
        DJANGO_HAWK={
            "HAWK_INCOMING_ACCESS_KEY": "some-id",
            "HAWK_INCOMING_SECRET_KEY": "some-secret",
        }
    )
    def test_seen_nonce(self):
        """
        Test if reusing a nonce within 60 seconds is rejected
        """

        url = self.get_url()
        sender = hawk_auth_sender(url=url)
        response = self.client.get(
            url,
            content_type="",
            HTTP_AUTHORIZATION=sender.request_header,
            HTTP_X_FORWARDED_FOR="1.2.3.4, 123.123.123.123",
        )

        self.assertEqual(response.status_code, 200)
        self.assertTrue("Content-Type" in response)
        self.assertTrue("Server-Authorization" in response)

        repeat_response = self.client.get(
            url,
            content_type="",
            HTTP_AUTHORIZATION=sender.request_header,
            HTTP_X_FORWARDED_FOR="1.2.3.4, 123.123.123.123",
        )
        self.assertEqual(repeat_response.status_code, 401)
        self.assertTrue("Content-Type" in response)
        self.assertTrue("Server-Authorization" in response)

    @override_settings(
        DJANGO_HAWK={
            "HAWK_INCOMING_ACCESS_KEY": "wrong-id",
            "HAWK_INCOMING_SECRET_KEY": "some-secret",
        }
    )
    def test_bad_credentials_mean_401_returned(self):
        """
        If the wrong credentials are used,
        then a 401 is returned
        """

        url = self.get_url()
        sender = hawk_auth_sender(url=url)
        response = self.client.get(
            url,
            content_type="",
            HTTP_AUTHORIZATION=sender.request_header,
            HTTP_X_FORWARDED_FOR="1.2.3.4, 123.123.123.123",
        )

        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response.json(),
            {
                "detail": "Incorrect authentication credentials.",
            },
        )
        self.assertTrue("Content-Type" in response)
        self.assertTrue("Server-Authorization" not in response)

    @override_settings(
        DJANGO_HAWK={
            "HAWK_INCOMING_ACCESS_KEY": "some-id",
            "HAWK_INCOMING_SECRET_KEY": "some-secret",
        }
    )
    def test_if_61_seconds_in_past_401_returned(self):
        """
        If the Authorization header is generated 61 seconds in the past, then a
        401 is returned
        """

        url = self.get_url()
        past = datetime.datetime.now() - datetime.timedelta(seconds=61)
        with freeze_time(past):
            auth = hawk_auth_sender(url).request_header
        response = self.client.get(
            url,
            content_type="",
            HTTP_AUTHORIZATION=auth,
            HTTP_X_FORWARDED_FOR="1.2.3.4, 123.123.123.123",
        )

        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response.json(),
            {
                "detail": "Incorrect authentication credentials.",
            },
        )
        self.assertTrue("Content-Type" in response)
        self.assertTrue("Server-Authorization" not in response)



from typing import Callable, Optional

from django.http import HttpRequest, HttpResponse
from django.utils.deprecation import MiddlewareMixin
from mohawk import Receiver

from django_hawk.settings import django_hawk_settings


class HawkResponseMiddleware(MiddlewareMixin):
    def __init__(self, get_response: Optional[Callable] = None) -> None:
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse:
        if not self.get_response:
            raise Exception("get_response is not defined")

        response = self.get_response(request)

        if self.is_hawk_request(request) and not response.has_header(
            "Server-Authorization"
        ):
            hawk_receiver = self.get_receiver(request)
            if hawk_receiver:
                response["Server-Authorization"] = hawk_receiver.respond(
                    content=response.content,
                    content_type=response["Content-Type"],
                )

        return response

    def is_hawk_request(self, request: HttpRequest) -> bool:
        if "HTTP_AUTHORIZATION" not in request.META:
            return False
        return request.META.get("HTTP_AUTHORIZATION", "").startswith("Hawk ")

    def get_receiver(self, request: HttpRequest) -> Optional[Receiver]:
        hawk_receiver = getattr(request, django_hawk_settings.REQUEST_ATTR_NAME, None)
        if isinstance(hawk_receiver, Receiver):
            return hawk_receiver
        return None



import logging

from django.http import HttpRequest, HttpResponse, JsonResponse

from django_hawk.utils import DjangoHawkAuthenticationFailed, authenticate_request

logger = logging.getLogger(__name__)


def simple_view(request: HttpRequest) -> HttpResponse:
    try:
        authenticate_request(request=request)
    except DjangoHawkAuthenticationFailed as e:
        logger.info("Authentication failed", e)
        return JsonResponse(
            status=401,
            data={
                "detail": e.message,
            },
        )
    return HttpResponse("This is a simple view")


def custom_header_view(request: HttpRequest) -> HttpResponse:
    try:
        authenticate_request(request=request)
    except DjangoHawkAuthenticationFailed as e:
        logger.info("Authentication failed", e)
        return JsonResponse(
            status=401,
            data={
                "detail": e.message,
            },
        )

    response = HttpResponse("This is a view with a custom header")
    response["Server-Authorization"] = "custom-header-value"

    return response



import logging
from typing import TYPE_CHECKING

from django.http import HttpRequest
from mohawk.exc import HawkFail

from django_hawk.authentication import authorise
from django_hawk.settings import django_hawk_settings

if TYPE_CHECKING:
    from mohawk import Receiver

logger = logging.getLogger(__name__)


class DjangoHawkAuthenticationFailed(Exception):
    message: str
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.message = args[0]


def authenticate_request(request: HttpRequest) -> "Receiver":
    if "HTTP_AUTHORIZATION" not in request.META:
        raise DjangoHawkAuthenticationFailed(
            django_hawk_settings.NO_CREDENTIALS_MESSAGE
        )

    try:
        hawk_receiver = authorise(request)
    except HawkFail as e:
        logger.warning(
            "Failed authentication {e}".format(
                e=e,
            )
        )
        raise DjangoHawkAuthenticationFailed(
            django_hawk_settings.INCORRECT_CREDENTIALS_MESSAGE
        )
    setattr(request, django_hawk_settings.REQUEST_ATTR_NAME, hawk_receiver)
    return hawk_receiver



from django.urls import path

from django_hawk.tests.views import custom_header_view, simple_view

urlpatterns = [
    path(
        "test-simple-view/",
        simple_view,
        name="test_simple_view",
    ),
    path(
        "test-custom-header-view/",
        custom_header_view,
        name="custom_header_view",
    ),
]



import logging
from typing import TypedDict

from django.core.cache import cache
from django.http import HttpRequest
from django.utils.crypto import constant_time_compare
from mohawk import Receiver
from mohawk.exc import HawkFail

from django_hawk.settings import django_hawk_settings

logger = logging.getLogger(__name__)


class LookupCredentials(TypedDict):
    id: str
    key: str
    algorithm: str


def lookup_credentials(access_key_id) -> LookupCredentials:
    """
    Raises a HawkFail if the passed ID is not equal to
    """

    if not constant_time_compare(
        access_key_id,
        django_hawk_settings.HAWK_INCOMING_ACCESS_KEY,
    ):
        raise HawkFail(
            "No Hawk ID of {access_key_id}".format(
                access_key_id=access_key_id,
            )
        )

    lookup_creds: LookupCredentials = {
        "id": django_hawk_settings.HAWK_INCOMING_ACCESS_KEY,
        "key": django_hawk_settings.HAWK_INCOMING_SECRET_KEY,
        "algorithm": "sha256",
    }
    return lookup_creds


def seen_nonce(access_key_id, nonce, _) -> bool:
    """
    Returns if the passed access_key_id/nonce combination has been
    used within 60 seconds.
    """

    cache_key = "activity_stream:{access_key_id}:{nonce}".format(
        access_key_id=access_key_id,
        nonce=nonce,
    )

    # cache.add only adds key if it isn't present
    seen_cache_key = not cache.add(
        cache_key,
        True,
        timeout=60,
    )

    if seen_cache_key:
        logger.warning("Already seen nonce {nonce}".format(nonce=nonce))

    return seen_cache_key


def authorise(request: HttpRequest) -> Receiver:
    """
    Raises a HawkFail if the passed request cannot be authenticated
    """
    return Receiver(
        lookup_credentials,
        request.META["HTTP_AUTHORIZATION"],
        request.build_absolute_uri(),
        request.method,
        content=request.body,
        content_type=request.content_type,
        seen_nonce=seen_nonce,
    )



from django.conf import settings

DEFAULTS = {
    "NO_CREDENTIALS_MESSAGE": "Authentication credentials were not provided.",
    "INCORRECT_CREDENTIALS_MESSAGE": "Incorrect authentication credentials.",
    "MAX_PER_PAGE": 500,
    "REQUEST_ATTR_NAME": "django_hawk_receiver",
}


class DjangoHawkSettings:
    HAWK_INCOMING_ACCESS_KEY: str
    HAWK_INCOMING_SECRET_KEY: str
    NO_CREDENTIALS_MESSAGE: str
    INCORRECT_CREDENTIALS_MESSAGE: str
    MAX_PER_PAGE: int
    REQUEST_ATTR_NAME: str

    def __getattr__(self, attr):
        django_settings = getattr(settings, "DJANGO_HAWK", {})

        # Check if present in user settings
        if attr in django_settings:
            return django_settings[attr]

        # Check if present in defaults
        default_value = DEFAULTS.get(attr, None)
        if default_value is None and attr not in DEFAULTS:
            raise AttributeError(f"No value set for DJANGO_HAWK['{attr}']")
        return default_value


django_hawk_settings = DjangoHawkSettings()
